# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _client module.
"""


import pytest

from zhmcclient import Client, CpcManager, MetricsContextManager
from zhmcclient_mock import FakedSession


@pytest.mark.parametrize(
    "hmc_name, hmc_version, api_version", [
        ('fake-hmc', '2.13.1', '1.8'),
    ]
)
def test_client_initial_attrs(hmc_name, hmc_version, api_version):
    """Test initial attributes of Client."""

    session = FakedSession('fake-host', hmc_name, hmc_version, api_version)

    # Execute the code to be tested
    client = Client(session)

    assert client.session is session
    assert isinstance(client.cpcs, CpcManager)
    assert client.cpcs.session is session
    assert isinstance(client.metrics_contexts, MetricsContextManager)
    assert client.metrics_contexts.session is session
    assert client.metrics_contexts.client is client


@pytest.mark.parametrize(
    "hmc_name, hmc_version, api_version", [
        ('fake-hmc1', '2.13.1', '1.8'),
        ('fake-hmc2', '2.14.0', '2.20'),
    ]
)
def test_version_info(hmc_name, hmc_version, api_version):
    """All tests for Client.version_info()."""

    session = FakedSession('fake-host', hmc_name, hmc_version, api_version)

    # Client object under test
    client = Client(session)

    # Execute the code to be tested
    version_info = client.version_info()

    exp_version_info = tuple([int(v) for v in api_version.split('.')])

    assert version_info == exp_version_info


@pytest.mark.parametrize(
    "hmc_name, hmc_version, api_version", [
        ('fake-hmc1', '2.13.1', '1.8'),
        ('fake-hmc2', '2.14.0', '2.20'),
    ]
)
def test_query_api_version(hmc_name, hmc_version, api_version):
    """All tests for Client.query_api_version()."""

    session = FakedSession('fake-host', hmc_name, hmc_version, api_version)

    # Client object under test
    client = Client(session)

    # Execute the code to be tested
    api_version_info = client.query_api_version()

    api_major_version = int(api_version.split('.')[0])
    api_minor_version = int(api_version.split('.')[1])

    exp_api_version_info = {
        'api-major-version': api_major_version,
        'api-minor-version': api_minor_version,
        'hmc-version': hmc_version,
        'hmc-name': hmc_name,
    }

    assert api_version_info == exp_api_version_info


TESTCASES_GET_INVENTORY = [
    # Testcases for test_get_inventory()
    # Each item in the list is a testcase with these properties:
    # - input_resources: Input parameter 'resources'.
    # - exp_exc_type: Expected exception type, or None.
    # - exp_inventory: Expected result inventory, as a dict with:
    #   key: Resource class
    #   value: List of resource names
    (
        None,
        Exception,
        None
    ),
    (
        [],
        None,
        {}
    ),
    (
        ['cpc'],
        None,
        {
            'cpc': ['CPC1'],
        }
    ),
    (
        ['core-resources'],
        None,
        {
            'cpc': ['CPC1'],
        }
    ),
    (
        ['partition'],
        None,
        {}
    ),
]


@pytest.mark.parametrize(
    "input_resources, exp_exc_type, exp_inventory",
    TESTCASES_GET_INVENTORY
)
def test_get_inventory(input_resources, exp_exc_type, exp_inventory):
    """All tests for Client.get_inventory()."""

    session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')

    # Client object under test
    client = Client(session)

    # We set up a fixed faked HMC environment
    session.hmc.cpcs.add({
        'object-id': 'fake-cpc1-oid',
        # object-uri is set up automatically
        'parent': None,
        'class': 'cpc',
        'name': 'CPC1',
        'description': 'CPC #1 (classic mode)',
        'status': 'active',
        'dpm-enabled': False,
        'is-ensemble-member': False,
        'iml-mode': 'lpar',
    })

    if exp_exc_type:
        try:

            # Execute the code to be tested
            client.get_inventory(input_resources)

        except exp_exc_type:
            pass
    else:

        # Execute the code to be tested
        inventory = client.get_inventory(input_resources)

        # Go through actual result and check against expected result
        seen_names_by_class = {}  # Resource classes and names already seen
        for resource_props in inventory:

            assert 'class' in resource_props
            res_class = resource_props['class']

            assert 'name' in resource_props
            res_name = resource_props['name']

            assert res_class in exp_inventory
            exp_names = exp_inventory[res_class]

            assert res_name in exp_names

            if res_class not in seen_names_by_class:
                seen_names_by_class[res_class] = []
            seen_names_by_class[res_class].append(res_name)

        # Check if there are any expected names that have not been in the result
        for exp_class, exp_names in exp_inventory.items():
            assert exp_class in seen_names_by_class
            seen_names = seen_names_by_class[exp_class]
            for exp_name in exp_names:
                assert exp_name in seen_names



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for adapters (on CPCs in DPM mode).

These tests do not change any existing adapters, but create, modify
and delete Hipersocket adapters.
"""


import uuid
import warnings
import pdb
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs, all_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    standard_partition_props, runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic Adapter objects (e.g. find_by_name())
ADAPTER_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in Adapter objects returned by list() without full props
ADAPTER_LIST_PROPS = ['object-uri', 'name', 'adapter-id', 'adapter-family',
                      'type', 'status']

# Properties in Adapter objects for list(additional_properties)
ADAPTER_ADDITIONAL_PROPS = ['description', 'detected-card-type']

# Properties whose values can change between retrievals of Adapter objects
ADAPTER_VOLATILE_PROPS = []


def test_adapter_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the adapters to test with
        adapter_list = cpc.adapters.list()
        if not adapter_list:
            skip_warn(f"No adapters on CPC {cpc.name} managed by HMC {hd.host}")
        adapter_list = pick_test_resources(adapter_list)

        for adapter in adapter_list:
            print(f"Testing on CPC {cpc.name} with adapter {adapter.name!r}")
            runtest_find_list(
                session, cpc.adapters, adapter.name, 'name', 'object-uri',
                ADAPTER_VOLATILE_PROPS, ADAPTER_MINIMAL_PROPS,
                ADAPTER_LIST_PROPS, ADAPTER_ADDITIONAL_PROPS)


def test_adapter_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the adapters to test with
        adapter_list = cpc.adapters.list()
        if not adapter_list:
            skip_warn(f"No adapters on CPC {cpc.name} managed by HMC {hd.host}")
        adapter_list = pick_test_resources(adapter_list)

        for adapter in adapter_list:
            print(f"Testing on CPC {cpc.name} with adapter {adapter.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(adapter.manager, non_list_prop)


def test_adapter_hs_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a Hipersocket adapter.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        print(f"Testing on CPC {cpc.name}")

        adapter_name = TEST_PREFIX + ' test_adapter_crud adapter1'
        adapter_name_new = adapter_name + ' new'

        # Ensure a clean starting point for this test
        try:
            adapter = cpc.adapters.find(name=adapter_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test Hipersocket adapter from previous run: "
                f"{adapter_name!r} on CPC {cpc.name}", UserWarning)
            adapter.delete()
        try:
            adapter = cpc.adapters.find(name=adapter_name_new)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test Hipersocket adapter from previous run: "
                f"{adapter_name_new!r} on CPC {cpc.name}", UserWarning)
            adapter.delete()

        # Create a Hipersocket adapter
        adapter_input_props = {
            'name': adapter_name,
            'description': 'Test adapter for zhmcclient end2end tests',
        }
        adapter_auto_props = {
            'type': 'hipersockets',
        }

        # The code to be tested
        adapter = cpc.adapters.create_hipersocket(adapter_input_props)

        for pn, exp_value in adapter_input_props.items():
            assert adapter.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        adapter.pull_full_properties()
        for pn, exp_value in adapter_input_props.items():
            assert adapter.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        for pn, exp_value in adapter_auto_props.items():
            assert adapter.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"

        # Test updating a property of the adapter

        new_desc = "Updated adapter description."

        # The code to be tested
        adapter.update_properties(dict(description=new_desc))

        assert adapter.properties['description'] == new_desc
        adapter.pull_full_properties()
        assert adapter.properties['description'] == new_desc

        # Test renaming the adapter

        # The code to be tested
        adapter.update_properties(dict(name=adapter_name_new))

        assert adapter.properties['name'] == adapter_name_new
        adapter.pull_full_properties()
        assert adapter.properties['name'] == adapter_name_new
        with pytest.raises(zhmcclient.NotFound):
            cpc.adapters.find(name=adapter_name)

        # Test deleting the adapter

        # The code to be tested
        adapter.delete()

        with pytest.raises(zhmcclient.NotFound):
            cpc.adapters.find(name=adapter_name_new)


def test_adapter_list_assigned_part(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test Adapter.list_assigned_partitions().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        required_families = [
            'hipersockets',
            'osa',
            'roce',
            'cna',
            'ficon',
            'accelerator',
            'crypto',
            'nvme',
            'coupling',
            'ism',
            'zhyperlink',
        ]
        family_adapters = {}
        for family in required_families:
            family_adapters[family] = []

        # List all adapters on the CPC and group by family
        adapter_list = cpc.adapters.list()
        for adapter in adapter_list:
            family = adapter.get_property('adapter-family')
            if family not in family_adapters:
                warnings.warn(
                    f"Ignoring adapter {adapter.name!r} on CPC {cpc.name} with "
                    f"an unknown family: '{family}'", UserWarning)
                continue
            if adapter.get_property('type') == 'osm':
                # Skip OSM adapters since they cannot be assigned to partitions
                continue
            family_adapters[family].append(adapter)

        tmp_part = None
        try:

            # Create a temporary partition for test purposes
            part_name = f"{TEST_PREFIX}_{uuid.uuid4().hex}"
            part_props = standard_partition_props(cpc, part_name)
            tmp_part = cpc.partitions.create(part_props)

            for family, all_adapters in family_adapters.items():
                if not all_adapters:
                    warnings.warn(
                        f"CPC {cpc.name} has no adapters of family '{family}'",
                        UserWarning)
                    continue

                if family in ('hipersockets', 'osa'):

                    test_adapters = pick_test_resources(all_adapters)
                    for test_adapter in test_adapters:
                        print(f"Testing on CPC {cpc.name} with adapter "
                              f"{test_adapter.name!r} (family '{family}')")

                        # The method to be tested
                        before_parts = test_adapter.list_assigned_partitions()

                        # Find the virtual switch for the test adapter
                        filter_args = {'backing-adapter-uri': test_adapter.uri}
                        vswitches = cpc.virtual_switches.list(
                            filter_args=filter_args)
                        assert len(vswitches) >= 1
                        vswitch = vswitches[0]

                        # Create a NIC in the temporary partition
                        nic_name = f"{family}_{uuid.uuid4().hex}"
                        nic_props = {
                            'name': nic_name,
                            'virtual-switch-uri': vswitch.uri,
                        }
                        tmp_part.nics.create(nic_props)

                        # The method to be tested
                        after_parts = test_adapter.list_assigned_partitions()

                        before_uris = [p.uri for p in before_parts]
                        new_parts = []
                        for part in after_parts:
                            if part.uri not in before_uris:
                                new_parts.append(part)

                        assert len(new_parts) == 1
                        new_part = new_parts[0]
                        assert new_part.uri == tmp_part.uri

                elif family in ('roce', 'cna'):

                    test_adapters = pick_test_resources(all_adapters)
                    for test_adapter in test_adapters:
                        print(f"Testing on CPC {cpc.name} with adapter "
                              f"{test_adapter.name!r} (family '{family}')")

                        # The method to be tested
                        before_parts = test_adapter.list_assigned_partitions()

                        test_port = test_adapter.ports.list()[0]

                        # Create a NIC in the temporary partition
                        nic_name = f"{family}_{uuid.uuid4().hex}"
                        nic_props = {
                            'name': nic_name,
                            'network-adapter-port-uri': test_port.uri,
                        }
                        tmp_part.nics.create(nic_props)

                        # The method to be tested
                        after_parts = test_adapter.list_assigned_partitions()

                        before_uris = [p.uri for p in before_parts]
                        new_parts = []
                        for part in after_parts:
                            if part.uri not in before_uris:
                                new_parts.append(part)

                        assert len(new_parts) == 1
                        new_part = new_parts[0]
                        assert new_part.uri == tmp_part.uri

                elif family == 'ficon':

                    # Assigning storage adapters to partitions is complex,
                    # so we search for a storage adapter that is already
                    # assigned to a partition.

                    found_test_adapter = False
                    for test_adapter in all_adapters:

                        if test_adapter.get_property('type') != 'fcp':
                            # This test works only for FCP adapters
                            continue

                        # The method to be tested
                        before_parts = test_adapter.list_assigned_partitions()

                        if len(before_parts) == 0:
                            continue

                        found_test_adapter = True
                        print(f"Testing on CPC {cpc.name} with FCP adapter "
                              f"{test_adapter.name!r} (family '{family}')")

                        found_adapters = []
                        for part in before_parts:
                            sgroups = part.list_attached_storage_groups()
                            for sg in sgroups:
                                if sg.get_property('type') != 'fcp':
                                    # This test works only for FCP stogrps
                                    continue
                                vsrs = sg.virtual_storage_resources.list()
                                for vsr in vsrs:
                                    port = vsr.adapter_port
                                    adapter = port.manager.parent
                                    if adapter.uri == test_adapter.uri:
                                        found_adapters.append(adapter)

                        assert len(found_adapters) >= 1
                        break

                    if not found_test_adapter:
                        warnings.warn(
                            f"CPC {cpc.name} has no FCP adapter with that is "
                            "assigned to any partition - cannot test FCP "
                            "adapters", UserWarning)

                elif family == 'accelerator':

                    test_adapters = pick_test_resources(all_adapters)
                    for test_adapter in test_adapters:
                        print(f"Testing on CPC {cpc.name} with adapter "
                              f"{test_adapter.name!r} (family '{family}')")

                        # The method to be tested
                        before_parts = test_adapter.list_assigned_partitions()

                        # Create a VF in the temporary partition
                        vf_name = f"{family}_{uuid.uuid4().hex}"
                        vf_props = {
                            'name': vf_name,
                            'adapter-uri': test_adapter.uri,
                        }
                        tmp_part.virtual_functions.create(vf_props)

                        # The method to be tested
                        after_parts = test_adapter.list_assigned_partitions()

                        before_uris = [p.uri for p in before_parts]
                        new_parts = []
                        for part in after_parts:
                            if part.uri not in before_uris:
                                new_parts.append(part)

                        assert len(new_parts) == 1
                        new_part = new_parts[0]
                        assert new_part.uri == tmp_part.uri

                elif family == 'crypto':

                    # Assigning crypto adapters to partitions is complex,
                    # so we search for a crypto adapter that is already
                    # assigned to a partition.

                    found_test_adapter = False
                    for test_adapter in all_adapters:

                        # The method to be tested
                        before_parts = test_adapter.list_assigned_partitions()

                        if len(before_parts) == 0:
                            continue

                        found_test_adapter = True
                        print(f"Testing on CPC {cpc.name} with adapter "
                              f"{test_adapter.name!r} (family '{family}')")

                        found_adapter_uris = []
                        for part in before_parts:
                            crypto_config = part.get_property(
                                'crypto-configuration')
                            if crypto_config is None:
                                continue
                            adapter_uris = crypto_config['crypto-adapter-uris']
                            for adapter_uri in adapter_uris:
                                if adapter_uri == test_adapter.uri:
                                    found_adapter_uris.append(adapter_uri)

                        assert len(found_adapter_uris) >= 1
                        break

                    if not found_test_adapter:
                        warnings.warn(
                            f"CPC {cpc.name} has no Crypto adapter that is "
                            "assigned to any partition - cannot test Crypto "
                            "adapters", UserWarning)

                elif family == 'nvme':
                    print(f"TODO: Implement test for family {family}")

                elif family == 'coupling':
                    print(f"TODO: Implement test for family {family}")

                elif family == 'ism':
                    print(f"TODO: Implement test for family {family}")

                elif family == 'zhyperlink':
                    print(f"TODO: Implement test for family {family}")

        finally:
            # Cleanup
            if tmp_part:
                tmp_part.delete()


def base_adapter_id(adapter_id, family):
    """
    Return the base adapter_id of an adapter_id.
    Both are in hex.
    """

    # Adapter families that have no physical adapter card
    virtual_families = [
        'hipersockets',
        'coupling',  # Internal Coupling Facility
        'ism',  # Internal Shared Memory
    ]

    pchid = int(adapter_id, 16)

    if family in virtual_families:

        # Make sure the virtual card detection based on the PCHID range in
        # Adapter.list_sibling_adapters() holds true
        assert pchid >= int('7c0', 16)

        # Virtual adapters have only 1 PCHID
        base_pchid = pchid

    else:

        # Make sure the physical card detection based on the PCHID range in
        # Adapter.list_sibling_adapters() holds true
        assert pchid < int('7c0', 16)

        # Physical adapters have 4 PCHIDs reeserved for the slot.
        # Calculate the base PCHID for the slot.
        base_pchid = pchid // 4 * 4

    return f'{base_pchid:03x}'


def test_adapter_list_sibling_adapters(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test Adapter.list_sibling_adapters().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        # Adapter families that have more than one Adapter object (=PCHID) on a
        # physical adapter card
        sibling_families = [
            'ficon',
            'crypto',
        ]

        adapters = cpc.adapters.list()

        # Determine the sibling adapters on the CPC, grouped by base PCHID
        adapters_by_base = {}  # key: base PCHID
        for adapter in adapters:
            adapter_id = adapter.get_property('adapter-id')
            family = adapter.get_property('adapter-family')
            if family in sibling_families:
                base_id = base_adapter_id(adapter_id, family)
                if base_id not in adapters_by_base:
                    adapters_by_base[base_id] = []
                adapters_by_base[base_id].append(adapter)

        # Test all adapters for their siblings
        for adapter in adapters:
            adapter_id = adapter.get_property('adapter-id')
            family = adapter.get_property('adapter-family')
            base_id = base_adapter_id(adapter_id, family)

            # The code to be tested
            sibling_adapters = adapter.list_sibling_adapters()

            assert isinstance(sibling_adapters, list)

            sibling_ids = {a.get_property('adapter-id')
                           for a in sibling_adapters}
            sibling_names = [a.name for a in sibling_adapters]
            if base_id not in adapters_by_base:
                exp_ids = set()
                exp_names = []
            else:
                exp_ids = {a.get_property('adapter-id')
                           for a in adapters_by_base[base_id]}
                exp_ids.remove(adapter_id)
                exp_names = [a.name for a in adapters_by_base[base_id]]
                exp_names.remove(adapter.name)

            assert sibling_ids == exp_ids, (
                f"Adapter '{adapter.name}' has unexpected siblings: got: "
                f"{sibling_names}, expected: {exp_names}")


# Properties in returned adapters, where each list item is a
# tuple(prop name, minimum CPC version).

# Default properties when no full/additional properties are requested:
DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS = [
    ('object-uri', [2, 13, 1]),
    ('name', [2, 13, 1]),
    ('adapter-id', [2, 13, 1]),
    ('adapter-family', [2, 13, 1]),
    ('type', [2, 13, 1]),
    ('status', [2, 13, 1]),
    # The 'firmware-update-pending' property requires the
    # LI_1580_CRYPTO_AUTO_TOGGLE feature to be enabled:
    ('firmware-update-pending', [2, 16, 0]),
    ('cpc-name', [2, 13, 1]),
    ('cpc-object-uri', [2, 13, 1]),
    ('se-version', [2, 13, 1]),
    ('dpm-enabled', [2, 13, 1]),
]

# Full set of properties that are common on all types of adapters:
COMMON_FULL_PROPS_ADAPTERS = DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS + [
    ('object-id', [2, 13, 1]),
    ('class', [2, 13, 1]),
    ('parent', [2, 13, 1]),
    ('description', [2, 13, 1]),
    ('detected-card-type', [2, 13, 1]),
    ('state', [2, 13, 1]),
    ('physical-channel-status', [2, 13, 1]),
]

LIST_PERMITTED_ADAPTERS_TESTCASES = [
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - input_kwargs (dict): Input parameters for the function.
    # - exp_props (list): List of expected properties in the result. Each list
    #     item is a tuple(prop name, minimum CPC version).
    # - exp_exc_type (class): Expected exception type, or None for success.
    # - run (bool or 'pdb'): Whether to run the test or call the debugger.

    (
        "Default parameters",
        dict(),
        DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS,
        None,
        True,
    ),
    (
        "full_properties",
        dict(
            full_properties=True,
        ),
        COMMON_FULL_PROPS_ADAPTERS,
        None,
        True,
    ),
    (
        "'detected-card-type' in additional_properties",
        dict(
            additional_properties=['detected-card-type'],
        ),
        DEFAULT_PROPS_LIST_PERMITTED_ADAPTERS + [
            ('detected-card-type', [2, 13, 1]),
        ],
        None,
        True,
    ),
]


@pytest.mark.parametrize(
    "desc, input_kwargs, exp_props, exp_exc_type, run",
    LIST_PERMITTED_ADAPTERS_TESTCASES)
def test_adapter_list_permitted(
        desc, input_kwargs, exp_props, exp_exc_type, run,
        all_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Test Console.list_permitted_adapters() without filtering, but with
    different variations of returned properties.
    """
    if not all_cpcs:
        pytest.skip("HMC definition does not include any CPCs")

    console = all_cpcs[0].manager.console
    session = console.manager.session
    client = console.manager.client
    hd = session.hmc_definition

    if hd.mock_file:
        skip_warn("zhmcclient mock does not support 'List Permitted Adapters' "
                  "operation")

    if run == 'pdb':
        # pylint: disable=forgotten-debug-statement
        pdb.set_trace()

    if not run:
        skip_warn("Testcase is disabled in testcase definition")

    if exp_exc_type:

        with pytest.raises(exp_exc_type) as exc_info:

            # Exercise the code to be tested
            _ = console.list_permitted_adapters(**input_kwargs)

        _ = exc_info.value
    else:

        # Exercise the code to be tested
        adapters = console.list_permitted_adapters(**input_kwargs)

        # Verify the result
        returned_cpc_names = set()
        for adapter in adapters:
            cpc = adapter.manager.parent
            returned_cpc_names.add(cpc.name)
            cpc_version = list(map(int, cpc.prop('se-version').split('.')))
            assert isinstance(adapter, zhmcclient.Adapter)
            for pname, min_cpc_version in exp_props:
                if cpc_version >= min_cpc_version:
                    # The property is supposed to be in the result
                    actual_pnames = list(adapter.properties.keys())
                    if pname == 'firmware-update-pending' \
                            and pname not in actual_pnames:
                        warnings.warn(
                            "The 'firmware-update-pending' property is not "
                            f"returned for adapters on CPC {cpc.name!r} (SE "
                            f"version {cpc.prop('se-version')}). Check on the "
                            "SE 'Manage Firmware Features' task whether "
                            "feature LI_1580_CRYPTO_AUTO_TOGGLE is enabled.")
                    else:
                        assert pname in actual_pnames, \
                            f"Actual adapter: {adapter!r}"

        # Verify that adapters are returned:
        # * For all DPM-mode CPCs (of any version)
        # * For all classic-mode CPCs of z15 or higher
        # The logic below assumes that each CPC has at least one adapter.
        expected_cpc_names = set()
        for cpc in client.cpcs.list():
            cpc_version = list(map(int, cpc.prop('se-version').split('.')))
            if cpc.dpm_enabled or cpc_version >= (2, 15):
                expected_cpc_names.add(cpc.name)
        assert returned_cpc_names == expected_cpc_names, (
            "Invalid set of CPCs of returned adapters:\n"
            f"Returned CPCs: {','.join(returned_cpc_names)}"
            f"Expected CPCs: {','.join(expected_cpc_names)}")



#!/usr/bin/env python
# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that performs an Increase Crypto Configuration operation on a partition
on a CPC in DPM mode.
"""

import sys
import getpass
import requests.packages.urllib3

import zhmcclient

requests.packages.urllib3.disable_warnings()

# HMC and user (password is prompted)
host = '10.11.12.13'
userid = 'myuser'
verify_cert = False

# CPC and partition to be updated
cpc_name = 'MYCPC'
partition_name = 'my_partition'

# Crypto adapters to be assigned to the partition
crypto_adapter_names = ['CRYP00']

# Crypto domains to be assigned to the partition and ther access mode
crypto_domains = range(2, 39)
access_mode = 'control'


def get_password(host, userid):
    prompt = f"Enter password for userid {userid} on HMC at {host}: "
    return getpass.getpass(prompt)


print(__doc__)

print(f"Using HMC at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, get_password=get_password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    try:
        print(f"Finding CPC {cpc_name} ...")
        cpc = client.cpcs.find(name=cpc_name)

        print(f"Finding partition {partition_name} ...")
        partition = cpc.partitions.find(name=partition_name)

        crypto_adapters = []
        for aname in crypto_adapter_names:
            print(f"Finding crypto adapter {aname} ...")
            adapter = cpc.adapters.find(name=aname)
            crypto_adapters.append(adapter)

        crypto_domain_config = []
        for domain in crypto_domains:
            domain_config = {
                "domain-index": domain,
                "access-mode": access_mode,
            }
            crypto_domain_config.append(domain_config)

        partition.increase_crypto_config(crypto_adapters, crypto_domain_config)

    except zhmcclient.Error as exc:
        print(f"Error: {exc}")
        sys.exit(1)

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for partitions (on CPCs in DPM mode).

These tests do not change any existing partitions, but create, modify and delete
test partitions.
"""


import warnings
import random
from datetime import timedelta, datetime, timezone
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    standard_partition_props, runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Print debug messages in tests
DEBUG = False

# Properties in minimalistic Partition objects (e.g. find_by_name())
PART_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in Partition objects returned by list() without full props
PART_LIST_PROPS = ['object-uri', 'name', 'status', 'type']

# Properties in Partition objects for list(additional_properties)
PART_ADDITIONAL_PROPS = ['description', 'short-name']

# Properties whose values can change between retrievals of Partition objects
PART_VOLATILE_PROPS = []


def test_part_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the partitions to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(f"No partitions on CPC {cpc.name} managed by "
                      f"HMC {hd.host}")
        part_list = pick_test_resources(part_list)

        for part in part_list:
            print(f"Testing on CPC {cpc.name} with partition {part.name!r}")
            runtest_find_list(
                session, cpc.partitions, part.name, 'name', 'status',
                PART_VOLATILE_PROPS, PART_MINIMAL_PROPS, PART_LIST_PROPS,
                PART_ADDITIONAL_PROPS)


def test_part_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the partitions to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(f"No partitions on CPC {cpc.name} managed by HMC "
                      f"{hd.host}")
        part_list = pick_test_resources(part_list)

        for part in part_list:
            print(f"Testing on CPC {cpc.name} with partition {part.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(part.manager, non_list_prop)


def test_part_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a partition.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        print(f"Testing on CPC {cpc.name}")

        part_name = TEST_PREFIX + ' test_part_crud part1'
        part_name_new = part_name + ' new'

        # Ensure a clean starting point for this test
        try:
            part = cpc.partitions.find(name=part_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test partition from previous run: {part_name!r} on "
                f"CPC {cpc.name}", UserWarning)
            status = part.get_property('status')
            if status != 'stopped':
                part.stop()
            part.delete()
        try:
            part = cpc.partitions.find(name=part_name_new)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test partition from previous run: {part_name_new!r} "
                f"on CPC {cpc.name}", UserWarning)
            status = part.get_property('status')
            if status != 'stopped':
                part.stop()
            part.delete()

        # Test creating the partition

        part_input_props = standard_partition_props(cpc, part_name)
        part_auto_props = {
            'status': 'stopped',
        }

        # The code to be tested
        part = cpc.partitions.create(part_input_props)

        for pn, exp_value in part_input_props.items():
            assert part.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        part.pull_full_properties()
        for pn, exp_value in part_input_props.items():
            assert part.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"
        for pn, exp_value in part_auto_props.items():
            assert part.properties[pn] == exp_value, \
                f"Unexpected value for property {pn!r}"

        # Test updating a property of the partition

        new_desc = "Updated partition description."

        # The code to be tested
        part.update_properties(dict(description=new_desc))

        assert part.properties['description'] == new_desc
        part.pull_full_properties()
        assert part.properties['description'] == new_desc

        # Test renaming the partition

        # The code to be tested
        part.update_properties(dict(name=part_name_new))

        assert part.properties['name'] == part_name_new
        part.pull_full_properties()
        assert part.properties['name'] == part_name_new
        with pytest.raises(zhmcclient.NotFound):
            cpc.partitions.find(name=part_name)

        # Test deleting the partition

        # The code to be tested
        part.delete()

        with pytest.raises(zhmcclient.NotFound):
            cpc.partitions.find(name=part_name_new)


def test_part_list_os_messages(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test "List OS Messages" operation on partitions
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support 'List OS Messages' "
                      "operation")

        # Pick the partition to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host}")

        test_part = None
        for part in part_list:

            # Test: List all messages (without begin or end)
            try:
                if DEBUG:
                    print("Debug: Test: Listing OS messages of partition "
                          f"{part.name} with no begin/end")
                result = part.list_os_messages()
            except zhmcclient.HTTPError as exc:
                if exc.http_status == 409 and exc.reason == 332:
                    # Meaning: The messages interface for the partition is not
                    # available
                    if DEBUG:
                        print(f"Debug: Partition {part.name} cannot list OS "
                              "messages")
                    continue
                raise

            all_messages = result['os-messages']
            if len(all_messages) >= 3:
                test_part = part
                break

            if DEBUG:
                print(f"Debug: Partition {part.name} has only "
                      f"{len(all_messages)} OS messages")

        if test_part is None:
            skip_warn(f"No partition on CPC {cpc.name} has the minimum number "
                      "of 3 OS messages for the test")

        # Test with begin/end selecting the full set of messages
        all_begin = all_messages[0]['sequence-number']
        all_end = all_messages[-1]['sequence-number']
        if DEBUG:
            print("Debug: Test: Listing OS messages of partition "
                  f"{test_part.name} with begin={all_begin}, end={all_end}")
        result = test_part.list_os_messages(begin=all_begin, end=all_end)
        messages = result['os-messages']
        assert len(messages) == all_end - all_begin + 1
        assert messages == all_messages

        # Test with begin/end selecting a subset of messages
        while True:
            seq1 = random.choice(all_messages)['sequence-number']
            if seq1 not in {all_begin, all_end}:
                break
        while True:
            seq2 = random.choice(all_messages)['sequence-number']
            if seq2 not in {all_begin, all_end, seq1}:
                break
        begin = min(seq1, seq2)
        end = max(seq1, seq2)
        if DEBUG:
            print("Debug: Test: Listing OS messages of partition "
                  f"{test_part.name} with begin={begin}, end={end}")
        result = test_part.list_os_messages(begin=begin, end=end)
        messages = result['os-messages']
        assert len(messages) == end - begin + 1
        for message in messages:
            assert message in all_messages


def test_part_create_os_websocket(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test "Get ASCII Console WebSocket URI" operation on partitions
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support 'Get ASCII Console "
                      "WebSocket URI' operation")

        # Pick the partition to test with
        active_part_list = cpc.partitions.list(filter_args={'status': 'active'})
        if not active_part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host} "
                "with an active status")

        # Pick a random partition to test with
        part = random.choice(active_part_list)

        ws_uri = part.create_os_websocket()

        assert ws_uri.startswith('/api/websock/')


# Full set of properties that are common on all types of partitions:
COMMON_PROPS_LIST = ['name', 'object-uri', 'type', 'status',
                     'has-unacceptable-status']
FULL_PROPS_LIST = COMMON_PROPS_LIST + [
    'initial-memory', 'maximum-memory', 'partition-id', 'processor-mode',
    'autogenerate-partition-id', 'boot-device', 'boot-network-device',
    'description', 'hba-uris', 'ifl-processors',
    'initial-ifl-processing-weight', 'nic-uris', 'partition-link-uris',
    'storage-group-uris', 'tape-link-uris', 'virtual-function-uris'
]
LIST_PERMITTED_PARTITION_TESTCASES = [
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - input_kwargs (dict): Input parameters for the function.
    # - exp_props (List) : Expected properties in the output
    (
        "Default parameters",
        dict(),
        COMMON_PROPS_LIST,
    ),
    (
        "full_properties",
        dict(
            full_properties=True,
        ),
        FULL_PROPS_LIST
    ),
    (
        "full_properties and filtering",
        dict(
            full_properties=True,
            filter_args={'type': 'linux'},
        ),
        FULL_PROPS_LIST
    ),
    (
        "Bad request in filter_args",
        dict(
            filter_args={'name': '@#1c'}
        ),
        COMMON_PROPS_LIST
    ),
    (
        "additional-properties",
        dict(
            additional_properties=['partition-id', 'tape-link-uris']
        ),
        COMMON_PROPS_LIST + ['partition-id', 'tape-link-uris']
    ),
]


@pytest.mark.parametrize(
    "desc, input_kwargs, exp_props",
    LIST_PERMITTED_PARTITION_TESTCASES)
def test_console_list_permitted_partitions(desc, input_kwargs, exp_props,
                                           dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name, unused-argument
    """
    Test list permitted partitions on a cpc
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        session = cpc.manager.session
        hd = session.hmc_definition
        client = zhmcclient.Client(session)
        console = client.consoles.console
        features = console.list_api_features()
        if 'additional_properties' in input_kwargs and \
                'dpm-ctc-partition-link-management' not in features and \
                'dpm-hipersockets-partition-link-management' not in features:
            pytest.skip("HMC does not support additional-properties parameter.")

        permitted_part_list = console.list_permitted_partitions(**input_kwargs)
        if not permitted_part_list:
            skip_warn(f"No partitions on CPC {cpc.name} managed by HMC "
                      f"{hd.host} for the user {session.userid}")

        permitted_part_list = pick_test_resources(permitted_part_list)
        for partition in permitted_part_list:
            assert isinstance(partition, zhmcclient.Partition)
            for pname in exp_props:
                # The property is supposed to be in the result
                actual_pnames = list(partition.properties.keys())
                assert pname in actual_pnames, \
                    f"Actual partition: {partition!r}"

        # Test list permitted partitions with filtering
        permitted_partitions_filter_agrs = (console.list_permitted_partitions(
            filter_args={'name': permitted_part_list[0].name,
                         'cpc-name': permitted_part_list[
                             0].manager.parent.name}))
        assert len(permitted_partitions_filter_agrs) == 1

        # Test list permitted partitions with bad filter args
        permitted_partitions_filter_agrs = (console.list_permitted_partitions(
            filter_args={'name': permitted_part_list[0].name,
                         'cpc-name': 'bad-cpc'}))
        assert len(permitted_partitions_filter_agrs) == 0


TESTCASES_PART_GET_SUSTAINABILITY_DATA = [
    # Test cases for test_part_get_sustainability_data(), each as a tuple with
    # these items:
    # * tc: Testcase short description
    # * input_kwargs: kwargs to be used as input parameters for
    #   Partition.get_sustainability_data()
    # * exp_oldest: expected delta time from now to oldest data point,
    #   as timedelta
    # * exp_delta: expected delta time between data points, as timedelta
    (
        "Default values (range=last-week, resolution=one-hour)",
        {},
        timedelta(days=7),
        timedelta(hours=1),
    ),
    (
        "range=last-day, resolution=one-hour",
        {
            "range": "last-day",
            "resolution": "one-hour",
        },
        timedelta(hours=24),
        timedelta(hours=1),
    ),
    (
        "range=last-day, resolution=fifteen-minutes",
        {
            "range": "last-day",
            "resolution": "fifteen-minutes",
        },
        timedelta(hours=24),
        timedelta(minutes=15),
    ),
]

PART_METRICS = {
    # Metrics returned in "Get Partition Historical Sustainability Data" HMC
    # operation.
    # metric name: data type
    "wattage": int,
    "processor-utilization": int,
}


@pytest.mark.parametrize(
    "tc, input_kwargs, exp_oldest, exp_delta",
    TESTCASES_PART_GET_SUSTAINABILITY_DATA)
def test_part_get_sustainability_data(
        tc, input_kwargs, exp_oldest, exp_delta,
        dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name,unused-argument
    """
    Test for Partition.get_sustainability_data(...)
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        if hd.mock_file:
            skip_warn("zhmcclient mock does not support "
                      "Partition.get_sustainability_data()")

        # Pick the partition to test with
        part_list = cpc.partitions.list()
        if not part_list:
            skip_warn(
                f"No partitions on CPC {cpc.name} managed by HMC {hd.host}")

        # Pick a random partition to test with
        part = random.choice(part_list)

        session = part.manager.session
        hd = session.hmc_definition

        print(f"Testing with partition {part.name}")

        try:

            # The code to be tested
            data = part.get_sustainability_data(**input_kwargs)

        except zhmcclient.HTTPError as exc:
            if exc.http_status == 403 and exc.reason == 1:
                skip_warn(
                    f"HMC userid {hd.userid!r} is not authorized for task "
                    f"'Environmental Dashboard' on HMC {hd.host}")
            elif exc.http_status == 404 and exc.reason == 1:
                skip_warn(
                    f"Partition {part.name} on HMC {hd.host} does not support "
                    f"feature: {exc}")
            else:
                raise

        now_dt = datetime.now(timezone.utc)
        exp_oldest_dt = now_dt - exp_oldest

        act_metric_names = set(data.keys())
        exp_metric_names = set(PART_METRICS.keys())
        assert act_metric_names == exp_metric_names

        for metric_name, metric_array in data.items():
            metric_type = PART_METRICS[metric_name]

            first_item = True
            previous_dt = None
            for dp_item in metric_array:
                # We assume the order is oldest to newest

                assert 'data' in dp_item
                assert 'timestamp' in dp_item
                assert len(dp_item) == 2

                dp_data = dp_item['data']
                dp_timestamp_dt = dp_item['timestamp']

                assert isinstance(dp_data, metric_type), \
                    f"Invalid data type for metric {metric_name!r}"

                if first_item:
                    first_item = False

                    # Verify that the oldest timestamp is within a certain
                    # delta from the range start.
                    # There are cases where that is not satisfied, so we only
                    # issue only a warning (as opposed to failing).
                    delta_sec = abs((dp_timestamp_dt - exp_oldest_dt).seconds)
                    if delta_sec > 15 * 60:
                        print(f"Warning: Oldest data point of metric "
                              f"{metric_name!r} is not within 15 minutes of "
                              "range start: "
                              f"Oldest data point: {dp_timestamp_dt}, "
                              f"Range start: {exp_oldest_dt}, "
                              f"Delta: {delta_sec} sec")
                else:

                    # For second oldest timestamp on, verify that the delta
                    # to the previous data point is the requested resolution.
                    # There are cases where that is not satisfied, so we only
                    # issue only a warning (as opposed to failing).
                    tolerance_pct = 10
                    delta_td = abs(dp_timestamp_dt - previous_dt)
                    if abs(delta_td.seconds - exp_delta.seconds) > \
                            tolerance_pct / 100 * exp_delta.seconds:
                        print("Warning: Timestamp of a data point of metric "
                              f"{metric_name!r} is not within expected delta "
                              "of its previous data point. "
                              f"Actual delta: {delta_td}, "
                              f"Expected delta: {exp_delta} "
                              f"(+/-{tolerance_pct}%)")

                previous_dt = dp_timestamp_dt



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example unit test for a user of the zhmcclient package.
"""


from requests.packages import urllib3

import zhmcclient
import zhmcclient_mock

urllib3.disable_warnings()


def create_session_1():
    """
    Demonstrate how to populate a faked session with resources defined
    in a resource dictionary.
    """
    session = zhmcclient_mock.FakedSession('fake-host', 'fake-hmc',
                                           '2.13.1', '1.8')
    session.hmc.add_resources({
        'cpcs': [
            {
                'properties': {
                    # object-id is auto-generated
                    # object-uri is auto-generated
                    'name': 'cpc_1',
                    'dpm-enabled': False,
                    'description': 'CPC #1',
                },
                'lpars': [
                    {
                        'properties': {
                            # object-id is auto-generated
                            # object-uri is auto-generated
                            'name': 'lpar_1',
                            'description': 'LPAR #1 in CPC #1',
                        },
                    },
                ],
            },
            {
                'properties': {
                    # object-id is auto-generated
                    # object-uri is auto-generated
                    'name': 'cpc_2',
                    'dpm-enabled': True,
                    'description': 'CPC #2',
                },
                'partitions': [
                    {
                        'properties': {
                            # object-id is auto-generated
                            # object-uri is auto-generated
                            'name': 'partition_1',
                            'description': 'Partition #1 in CPC #2',
                        },
                    },
                ],
                'adapters': [
                    {
                        'properties': {
                            # object-id is auto-generated
                            # object-uri is auto-generated
                            'name': 'osa_1',
                            'description': 'OSA #1 in CPC #2',
                            'type': 'osd',
                        },
                        'ports': [
                            {
                                'properties': {
                                    # element-id is auto-generated
                                    # element-uri is auto-generated
                                    'name': 'osa_1_port_1',
                                    'description': 'Port #1 of OSA #1',
                                },
                            },
                        ],
                    },
                ],
            },
        ],
    })
    return session


def create_session_2():
    """
    Demonstrate how to populate a faked session with resources one by one.
    """
    session = zhmcclient_mock.FakedSession('fake-host', 'fake-hmc',
                                           '2.13.1', '1.8')
    cpc1 = session.hmc.cpcs.add({
        # object-id is auto-generated
        # object-uri is auto-generated
        'name': 'cpc_1',
        'dpm-enabled': False,
        'description': 'CPC #1',
    })
    cpc1.lpars.add({
        # object-id is auto-generated
        # object-uri is auto-generated
        'name': 'lpar_1',
        'description': 'LPAR #1 in CPC #1',
    })
    cpc2 = session.hmc.cpcs.add({
        # object-id is auto-generated
        # object-uri is auto-generated
        'name': 'cpc_2',
        'dpm-enabled': True,
        'description': 'CPC #2',
    })
    cpc2.partitions.add({
        # object-id is auto-generated
        # object-uri is auto-generated
        'name': 'partition_1',
        'description': 'Partition #1 in CPC #2',
    })
    adapter1 = cpc2.adapters.add({
        # object-id is auto-generated
        # object-uri is auto-generated
        'name': 'osa_1',
        'description': 'OSA #1 in CPC #2',
        'type': 'osd',
    })
    adapter1.ports.add({
        # element-id is auto-generated
        # element-uri is auto-generated
        'name': 'osa_1_port_1',
        'description': 'Port #1 of OSA #1',
    })
    return session


def check(session, client):
    """
    Check the faked session and its faked HMC.
    """

    assert session.host == 'fake-host'

    assert client.version_info() == (1, 8)

    cpcs = client.cpcs.list()
    assert len(cpcs) == 2

    cpc1 = cpcs[0]  # a CPC in classic mode
    assert cpc1.get_property('name') == 'cpc_1'
    assert not cpc1.dpm_enabled

    lpars = cpc1.lpars.list()
    assert len(lpars) == 1
    lpar1 = lpars[0]
    assert lpar1.get_property('name') == 'lpar_1'

    cpc2 = cpcs[1]  # a CPC in DPM mode
    assert cpc2.get_property('name') == 'cpc_2'
    assert cpc2.dpm_enabled

    partitions = cpc2.partitions.list()
    assert len(partitions) == 1
    partition1 = partitions[0]
    assert partition1.get_property('name') == 'partition_1'

    adapters = cpc2.adapters.list()
    assert len(adapters) == 1
    adapter1 = adapters[0]
    assert adapter1.get_property('name') == 'osa_1'

    ports = adapter1.ports.list()
    assert len(ports) == 1
    port1 = ports[0]
    assert port1.get_property('name') == 'osa_1_port_1'


def test_session_1():
    """
    Test function 1
    """
    session = create_session_1()
    client = zhmcclient.Client(session)
    check(session, client)


def test_session_2():
    """
    Test function 2
    """
    session = create_session_2()
    client = zhmcclient.Client(session)
    check(session, client)



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _ldap_srv_def module.
"""


import re
import copy
import logging
import pytest

from zhmcclient import Client, HTTPError, NotFound, LdapServerDefinition
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources, assert_blanked_in_message


class TestLdapServerDefinition:
    """All tests for the LdapServerDefinition and LdapServerDefinitionManager
    classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked Console without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_console = self.session.hmc.consoles.add({
            'object-id': None,
            # object-uri will be automatically set
            'parent': None,
            'class': 'console',
            'name': 'fake-console1',
            'description': 'Console #1',
        })
        self.console = self.client.consoles.find(name=self.faked_console.name)

    def add_ldap_srv_def(self, name):
        """
        Add a faked LDAPServerDefinition object to the faked Console
        and return it.
        """

        faked_ldap_srv_def = self.faked_console.ldap_server_definitions.add({
            'element-id': f'oid-{name}',
            # element-uri will be automatically set
            'parent': '/api/console',
            'class': 'ldap-server-definition',
            'name': name,
            'description': f'LDAP Server Definition {name}',
            'primary-hostname-ipaddr': f'host-{name}',
        })
        return faked_ldap_srv_def

    def test_ldap_srv_def_manager_repr(self):
        """Test LdapServerDefinitionManager.__repr__()."""

        ldap_srv_def_mgr = self.console.ldap_server_definitions

        # Execute the code to be tested
        repr_str = repr(ldap_srv_def_mgr)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{ldap_srv_def_mgr.__class__.__name__}\s+at\s+'
            rf'0x{id(ldap_srv_def_mgr):08x}\s+\(\\n.*',
            repr_str)

    def test_ldap_srv_def_manager_initial_attrs(self):
        """Test initial attributes of LdapServerDefinitionManager."""

        ldap_srv_def_mgr = self.console.ldap_server_definitions

        # Verify all public properties of the manager object
        assert ldap_srv_def_mgr.resource_class == LdapServerDefinition
        assert ldap_srv_def_mgr.class_name == 'ldap-server-definition'
        assert ldap_srv_def_mgr.session is self.session
        assert ldap_srv_def_mgr.parent is self.console
        assert ldap_srv_def_mgr.console is self.console

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            (dict(full_properties=False),
             ['element-uri', 'name']),
            (dict(full_properties=True),
             ['element-uri', 'name', 'description']),
            ({},  # test default for full_properties (False)
             ['element-uri', 'name']),
        ]
    )
    @pytest.mark.parametrize(
        "filter_args, exp_names", [
            (None,
             ['a', 'b']),
            ({},
             ['a', 'b']),
            ({'name': 'a'},
             ['a']),
            ({'name': 'A'},  # LDAP user definitions have case-insensitive names
             ['a']),
        ]
    )
    def test_ldap_srv_def_manager_list(
            self, filter_args, exp_names, full_properties_kwargs, prop_names):
        """Test LdapServerDefinitionManager.list()."""

        faked_ldap_srv_def1 = self.add_ldap_srv_def(name='a')
        faked_ldap_srv_def2 = self.add_ldap_srv_def(name='b')
        faked_ldap_srv_defs = [faked_ldap_srv_def1, faked_ldap_srv_def2]
        exp_faked_ldap_srv_defs = [u for u in faked_ldap_srv_defs
                                   if u.name in exp_names]
        ldap_srv_def_mgr = self.console.ldap_server_definitions

        # Execute the code to be tested
        ldap_srv_defs = ldap_srv_def_mgr.list(filter_args=filter_args,
                                              **full_properties_kwargs)

        assert_resources(ldap_srv_defs, exp_faked_ldap_srv_defs, prop_names)

    @pytest.mark.parametrize(
        "input_props, exp_prop_names, exp_exc", [
            ({},  # props missing
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'description': 'fake description X'},  # props missing
             None,
             HTTPError({'http-status': 400, 'reason': 5})),
            ({'description': 'fake description X',
              'name': 'a',
              'primary-hostname-ipaddr': '10.11.12.13',
              'search-distinguished-name': 'test{0}'},
             ['element-uri', 'name', 'description'],
             None),
            ({'name': 'a',
              'primary-hostname-ipaddr': '10.11.12.13',
              'search-distinguished-name': 'test{0}',
              'bind-password': 'bla'},
             ['element-uri', 'name', 'bind-password'],
             None),
        ]
    )
    def test_ldap_srv_def_manager_create(
            self, caplog, input_props, exp_prop_names, exp_exc):
        """Test LdapServerDefinitionManager.create()."""

        logger_name = "zhmcclient.api"
        caplog.set_level(logging.DEBUG, logger=logger_name)

        ldap_srv_def_mgr = self.console.ldap_server_definitions

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                ldap_srv_def_mgr.create(properties=input_props)

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

        else:

            # Execute the code to be tested.
            ldap_srv_def = ldap_srv_def_mgr.create(properties=input_props)

            # Get its API call log record
            call_record = caplog.records[-2]

            # Check the resource for consistency within itself
            assert isinstance(ldap_srv_def, LdapServerDefinition)
            ldap_srv_def_name = ldap_srv_def.name
            exp_ldap_srv_def_name = ldap_srv_def.properties['name']
            assert ldap_srv_def_name == exp_ldap_srv_def_name
            ldap_srv_def_uri = ldap_srv_def.uri
            exp_ldap_srv_def_uri = ldap_srv_def.properties['element-uri']
            assert ldap_srv_def_uri == exp_ldap_srv_def_uri

            # Check the properties against the expected names and values
            for prop_name in exp_prop_names:
                assert prop_name in ldap_srv_def.properties
                if prop_name in input_props:
                    value = ldap_srv_def.properties[prop_name]
                    exp_value = input_props[prop_name]
                    assert value == exp_value

            # Verify the API call log record for blanked-out properties.
            assert_blanked_in_message(
                call_record.message, input_props,
                ['bind-password'])

    def test_ldap_srv_def_repr(self):
        """Test LdapServerDefinition.__repr__()."""

        faked_ldap_srv_def1 = self.add_ldap_srv_def(name='a')
        ldap_srv_def1 = self.console.ldap_server_definitions.find(
            name=faked_ldap_srv_def1.name)

        # Execute the code to be tested
        repr_str = repr(ldap_srv_def1)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{ldap_srv_def1.__class__.__name__}\s+at\s+'
            rf'0x{id(ldap_srv_def1):08x}\s+\(\\n.*',
            repr_str)

    @pytest.mark.parametrize(
        "input_props, exp_exc", [
            ({'name': 'a'},
             None),
            ({'name': 'b'},
             None),
        ]
    )
    def test_ldap_srv_def_delete(self, input_props, exp_exc):
        """Test LdapServerDefinition.delete()."""

        faked_ldap_srv_def = self.add_ldap_srv_def(name=input_props['name'])

        ldap_srv_def_mgr = self.console.ldap_server_definitions
        ldap_srv_def = ldap_srv_def_mgr.find(name=faked_ldap_srv_def.name)

        if exp_exc is not None:

            with pytest.raises(exp_exc.__class__) as exc_info:

                # Execute the code to be tested
                ldap_srv_def.delete()

            exc = exc_info.value
            if isinstance(exp_exc, HTTPError):
                assert exc.http_status == exp_exc.http_status
                assert exc.reason == exp_exc.reason

            # Check that the LDAP Server Definition still exists
            ldap_srv_def_mgr.find(name=faked_ldap_srv_def.name)

        else:

            # Execute the code to be tested.
            ldap_srv_def.delete()

            # Check that the LDAP Server Definition no longer exists
            with pytest.raises(NotFound) as exc_info:
                ldap_srv_def_mgr.find(name=faked_ldap_srv_def.name)

    def test_ldap_delete_create_same(self):
        """Test LdapServerDefinition.delete() followed by create() with same
        name."""

        ldap_srv_def_name = 'faked_a'

        # Add the LDAP Server Definition to be tested
        self.add_ldap_srv_def(name=ldap_srv_def_name)

        # Input properties for a LDAP Server Definition with the same name
        sn_ldap_srv_def_props = {
            'name': ldap_srv_def_name,
            'description': 'LDAP Server Definition with same name',
            'primary-hostname-ipaddr': '10.11.12.13',
            'search-distinguished-name': 'test{0}',
        }

        ldap_srv_def_mgr = self.console.ldap_server_definitions
        ldap_srv_def = ldap_srv_def_mgr.find(name=ldap_srv_def_name)

        # Execute the deletion code to be tested
        ldap_srv_def.delete()

        # Check that the LDAP Server Definition no longer exists
        with pytest.raises(NotFound):
            ldap_srv_def_mgr.find(name=ldap_srv_def_name)

        # Execute the creation code to be tested.
        ldap_srv_def_mgr.create(sn_ldap_srv_def_props)

        # Check that the LDAP Server Definition exists again under that name
        sn_ldap_srv_def = ldap_srv_def_mgr.find(name=ldap_srv_def_name)
        description = sn_ldap_srv_def.get_property('description')
        assert description == sn_ldap_srv_def_props['description']

    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New LDAP Server Definition description'},
            {'bind-password': 'bla'},
        ]
    )
    def test_ldap_srv_def_update_properties(self, caplog, input_props):
        """Test LdapServerDefinition.update_properties()."""

        logger_name = "zhmcclient.api"
        caplog.set_level(logging.DEBUG, logger=logger_name)

        ldap_srv_def_name = 'faked_a'

        # Add the LDAP Server Definition to be tested
        self.add_ldap_srv_def(name=ldap_srv_def_name)

        ldap_srv_def_mgr = self.console.ldap_server_definitions
        ldap_srv_def = ldap_srv_def_mgr.find(name=ldap_srv_def_name)

        ldap_srv_def.pull_full_properties()
        saved_properties = copy.deepcopy(ldap_srv_def.properties)

        # Execute the code to be tested
        ldap_srv_def.update_properties(properties=input_props)

        # Get its API call log record
        call_record = caplog.records[-2]

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in ldap_srv_def.properties
            prop_value = ldap_srv_def.properties[prop_name]
            assert prop_value == exp_prop_value, \
                f"Unexpected value for property {prop_name!r}"

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        ldap_srv_def.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in ldap_srv_def.properties
            prop_value = ldap_srv_def.properties[prop_name]
            assert prop_value == exp_prop_value

        # Verify the API call log record for blanked-out properties.
        assert_blanked_in_message(
            call_record.message, input_props,
            ['bind-password'])



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Definition of the package version, and check for supported Python versions.
"""

import sys

# In the RTD docs build, _version_scm.py does not exist:
try:
    from ._version_scm import version, version_tuple
except ImportError:
    version = "unknown"
    version_tuple = tuple("unknown")

__all__ = ['__version__', '__version_tuple__']

#: The full version of this package including any development levels, as a
#: :term:`string`.
#:
#: Possible formats for this version string are:
#:
#: * "M.N.Pa1.dev7+g1234567": A not yet released version M.N.P
#: * "M.N.P": A released version M.N.P
__version__ = version

#: The full version of this package including any development levels, as a
#: tuple of version items, converted to integer where possible.
#:
#: Possible formats for this version string are:
#:
#: * (M, N, P, 'a1', 'dev7', 'g1234567'): A not yet released version M.N.P
#: * (M, N, P): A released version M.N.P
__version_tuple__ = version_tuple

# Check supported Python versions
# Keep these Python versions in sync with:
# - python_requires and classifiers in setup.py
# - Section "Supported environments" in docs/intro.rst
if sys.version_info[0:2] < (3, 8):
    raise RuntimeError('zhmcclient requires Python 3.8 or higher')



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for virtual functions (on CPCs in DPM mode).

These tests do not change any existing partitions or virtual functions, but
create, modify and delete test partitions with virtual functions.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    standard_partition_props, runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic VirtualFunction objects (e.g. find_by_name())
VFUNC_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in VirtualFunction objects returned by list() without full props
VFUNC_LIST_PROPS = ['element-uri', 'name', 'description']

# Properties whose values can change between retrievals of VirtualFunction
# objects
VFUNC_VOLATILE_PROPS = []


def test_vfunc_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the virtual functions to test with
        part_vfunc_tuples = []
        part_list = cpc.partitions.list()
        for part in part_list:
            vfunc_list = part.virtual_functions.list()
            for vfunc in vfunc_list:
                part_vfunc_tuples.append((part, vfunc))
        if not part_vfunc_tuples:
            skip_warn(
                f"No partitions with virtual functions on CPC {cpc.name} "
                f"managed by HMC {hd.host}")
        part_vfunc_tuples = pick_test_resources(part_vfunc_tuples)

        for part, vfunc in part_vfunc_tuples:
            print(f"Testing on CPC {cpc.name} with virtual function "
                  f"{vfunc.name!r} of partition {part.name!r}")
            runtest_find_list(
                session, part.virtual_functions, vfunc.name, 'name',
                'description', VFUNC_VOLATILE_PROPS, VFUNC_MINIMAL_PROPS,
                VFUNC_LIST_PROPS)


def test_vfunc_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the virtual functions to test with
        part_vfunc_tuples = []
        part_list = cpc.partitions.list()
        for part in part_list:
            vfunc_list = part.virtual_functions.list()
            for vfunc in vfunc_list:
                part_vfunc_tuples.append((part, vfunc))
        if not part_vfunc_tuples:
            skip_warn(
                f"No partitions with virtual functions on CPC {cpc.name} "
                f"managed by HMC {hd.host}")
        part_vfunc_tuples = pick_test_resources(part_vfunc_tuples)

        for part, vfunc in part_vfunc_tuples:
            print(f"Testing on CPC {cpc.name} with virtual function "
                  f"{vfunc.name!r} of partition {part.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(vfunc.manager, non_list_prop)


def test_vfunc_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a virtual function (and a partition).
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        print(f"Testing on CPC {cpc.name}")

        part_name = TEST_PREFIX + ' test_vfunc_crud part1'
        vfunc_name = 'vfunc1'
        vfunc_name_new = vfunc_name + ' new'

        # Ensure a clean starting point for this test
        try:
            part = cpc.partitions.find(name=part_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test partition from previous run: {part_name!r} "
                f"on CPC {cpc.name}", UserWarning)
            status = part.get_property('status')
            if status != 'stopped':
                part.stop()
            part.delete()

        part = None
        try:

            # Pick a zEDC accelerator adapter that will back the virtual func.
            edc_adapters = cpc.adapters.findall(type='zedc')
            if not edc_adapters:
                skip_warn(f"No zEDC accelerator adapters on CPC {cpc.name} "
                          f"managed by HMC {hd.host}")
            edc_adapter = edc_adapters[-1]  # Pick the last one

            # Create a partition that will lateron contain the virtual function
            part_props = standard_partition_props(cpc, part_name)
            part = cpc.partitions.create(part_props)

            # Test creating a virtual function

            vfunc_input_props = {
                'name': vfunc_name,
                'description': 'Dummy virtual function description.',
                'adapter-uri': edc_adapter.uri,
                'device-number': '0100',
            }
            vfunc_auto_props = {}

            # The code to be tested
            vfunc = part.virtual_functions.create(vfunc_input_props)

            for pn, exp_value in vfunc_input_props.items():
                assert vfunc.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"
            vfunc.pull_full_properties()
            for pn, exp_value in vfunc_input_props.items():
                assert vfunc.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"
            for pn, exp_value in vfunc_auto_props.items():
                assert vfunc.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"

            # Test updating a property of the virtual function

            new_desc = "Updated virtual function description."

            # The code to be tested
            vfunc.update_properties(dict(description=new_desc))

            assert vfunc.properties['description'] == new_desc
            vfunc.pull_full_properties()
            assert vfunc.properties['description'] == new_desc

            # Test renaming the virtual function

            # The code to be tested
            vfunc.update_properties(dict(name=vfunc_name_new))

            assert vfunc.properties['name'] == vfunc_name_new
            vfunc.pull_full_properties()
            assert vfunc.properties['name'] == vfunc_name_new
            with pytest.raises(zhmcclient.NotFound):
                part.virtual_functions.find(name=vfunc_name)

            # Test deleting the virtual function

            # The code to be tested
            vfunc.delete()

            with pytest.raises(zhmcclient.NotFound):
                part.virtual_functions.find(name=vfunc_name_new)

        finally:
            # Cleanup
            if part:
                part.delete()



#!/usr/bin/env python
#
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
#   Copyright Red Hat
#
#   SPDX-License-Identifier: GPL-2.0
#
#   Author: Sebastian Mitterle <smitterl@redhat.com>

"""
Example that triggers load from ftp from a specified generic.ins
and displays the console output.
This is useful when the .prm file has a kickstart reference for
a fully automated installation.
"""
import sys
import time
import requests.packages.urllib3
import yaml
import stomp
import zhmcclient

requests.packages.urllib3.disable_warnings()

PRINT_METADATA = False


def receive_until_KeyboardInterrupt(receiver):
    while True:
        try:
            for headers, message in receiver.notifications():
                os_msg_list = message['os-messages']
                for os_msg in os_msg_list:
                    if PRINT_METADATA:
                        msg_id = os_msg['message-id']
                        held = os_msg['is-held']
                        priority = os_msg['is-priority']
                        prompt = os_msg.get('prompt-text', None)
                        print(f"# OS message {msg_id} (held: {held}, "
                              f"priority: {priority}, prompt: {prompt}):")
                    msg_txt = os_msg['message-text'].strip('\n')
                    print(msg_txt)
        except zhmcclient.NotificationError as exc:
            print(f"Notification Error: {exc} - reconnecting")
            continue
        except stomp.exception.StompException as exc:
            print(f"STOMP Error: {exc} - reconnecting")
            continue
        except KeyboardInterrupt:
            print("Keyboard interrupt - leaving receiver loop")
            receiver.close()
            break
        else:
            raise AssertionError("Receiver was closed - should not happen")


def get_password(host, userid):
    return input(f"password for user {userid} on host {host}:")


def load_config(config_filepath):
    """
    config = {
        'host': "<host_url>",
        'userid': "<userid>",
        'ftp_config': {
            'host' : '<ftp_server_url>',
            'username' : '<ftp_user>',
            'password' : '<ftp_password>',
            'file_path' : '<ftp_file_path>'
        },
        'lpar': "<lpar-name>"
    }
    """
    with open(config_filepath) as f:
        return yaml.safe_load(f)


def main(config_filepath):
    config = load_config(config_filepath)
    session = zhmcclient.Session(config['host'], config['userid'], get_password=get_password, verify_cert=False)
    client = zhmcclient.Client(session)
    console = client.consoles.console
    lpars = console.list_permitted_lpars()
    lpar = [x for x in lpars if x.properties.get("name") == config['lpar']][0]
    print(lpar)
    ftp_config = config['ftp_config']
    lpar.load_from_ftp(host=ftp_config['host'], username=ftp_config['username'],
                       password=ftp_config['password'], load_file=ftp_config['file_path'],
                       wait_for_completion=False)
    print("Load from FTP issued. Will connect to OS messages in 10 sec.")
    time.sleep(10)
    topic = lpar.open_os_message_channel()
    receiver = zhmcclient.NotificationReceiver(topic, config['host'], session.session_id, session.session_credential)
    receive_until_KeyboardInterrupt(receiver)


if __name__ == '__main__':
    main(sys.argv[1])



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`User Role` resource represents an authority role which can be assigned
to one or more HMC users.

A User Role may allow access to specific managed objects, classes of managed
objects, groups and/or tasks. There are two types of User Roles: user-defined
and system-defined. User-defined User Roles are created by an HMC user, whereas
the system-defined User Roles are pre-defined, standard User Roles supplied
with the HMC.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_USER_ROLE

__all__ = ['UserRoleManager', 'UserRole']


class UserRoleManager(BaseManager):
    """
    Manager providing access to the :term:`User Role` resources of a HMC.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Console` object:

    * :attr:`zhmcclient.Console.user_roles`

    HMC/SE version requirements:

    * HMC version >= 2.13.0
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #      Console object representing the HMC.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        # Because this resource has case-insensitive names, this list must
        # contain the name property.
        query_props = [
            'name',
            'type',
        ]

        super().__init__(
            resource_class=UserRole,
            class_name=RC_USER_ROLE,
            session=console.manager.session,
            parent=console,
            base_uri='/api/user-roles',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props,
            case_insensitive_names=True)

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: :term:`Console` defining the scope for
        this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the :term:`User Role` resources representing the user roles
        defined in this HMC.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * User-related-access permission to the User Role objects included in
          the result, or task permission to the "Manage User Roles" task.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.UserRole` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'user-roles'
        list_uri = f'{self.console.uri}/user-roles'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties):
        """
        Create a new (user-defined) User Role in this HMC.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Roles" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create User Role' in the :term:`HMC API` book.

        Returns:

          UserRole:
            The resource object for the new User Role.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.console.uri + '/user-roles',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        user_role = UserRole(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return user_role


class UserRole(BaseResource):
    """
    Representation of a :term:`User Role`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.UserRoleManager`).

    HMC/SE version requirements:

    * HMC version >= 2.13.0
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.UserRoleManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, UserRoleManager), (
            f"Console init: Expected manager type {UserRoleManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this User Role.

        The User Role must be user-defined. System-defined User Roles cannot be
        deleted.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Roles" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this User Role.

        The User Role must be user-defined. System-defined User Roles cannot be
        updated.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Roles" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'User Role object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)

        # The name of User Roles cannot be updated. An attempt to do so should
        # cause HTTPError to be raised in the POST above, so we assert that
        # here, because we omit the extra code for handling name updates:
        assert self.manager._name_prop not in properties
        self.update_properties_local(copy.deepcopy(properties))

    @logged_api_call
    def add_permission(self, permitted_object, include_members=None,
                       view_only=None):
        # pylint: disable=line-too-long
        """
        Add permission for the specified permitted object(s) to this User Role,
        thereby granting that permission to all users that have this User Role.

        The granted permission depends on the resource class of the permitted
        object(s):

        * For Task resources, the granted permission is task permission for
          that task.

        * For Group resources, the granted permission is object access
          permission for the group resource, and optionally also for the
          group members.

        * For any other resources, the granted permission is object access
          permission for these resources.

        The User Role must be user-defined.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Roles" task.

        Parameters:

          permitted_object (:class:`~zhmcclient.BaseResource` or :term:`string`):
            Permitted object(s), either as a Python resource object (e.g.
            :class:`~zhmcclient.Partition`), or as a resource class string (e.g.
            'partition').

            Must not be `None`.

          include_members (bool): Controls whether for Group resources, the
            operation applies additionally to its group member resources.

            If `None`, this parameter will be ignored. It should only be
            specified when the permitted object is a Group resource.

          view_only (bool): Controls whether for Task resources, the operation
            aplies to the view-only version of the task (if `True`), or to
            the full version of the task (if `False`). Only certain tasks
            support a view-only version.

            If `None`, this parameter will be ignored. It should only be
            specified when the permitted object is a Task resource.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """  # noqa: E501
        # pylint: enable=line-too-long
        if isinstance(permitted_object, BaseResource):
            perm_obj = permitted_object.uri
            perm_type = 'object'
        elif isinstance(permitted_object, str):
            perm_obj = permitted_object
            perm_type = 'object-class'
        else:
            raise TypeError(
                "permitted_object must be a string or BaseResource, but is: "
                f"{type(permitted_object)}")
        body = {
            'permitted-object': perm_obj,
            'permitted-object-type': perm_type,
        }
        if include_members is not None:
            body['include-members'] = include_members
        if view_only is not None:
            body['view-only-mode'] = view_only
        self.manager.session.post(
            self.uri + '/operations/add-permission', resource=self,
            body=body)

    @logged_api_call
    def remove_permission(self, permitted_object, include_members=None,
                          view_only=None):
        # pylint: disable=line-too-long
        """
        Remove permission for the specified permitted object(s) from this User
        Role, thereby no longer granting that permission to all users that have
        this User Role.

        The granted permission depends on the resource class of the permitted
        object(s):

        * For Task resources, the granted permission is task permission for
          that task.

        * For Group resources, the granted permission is object access
          permission for the group resource, and optionally also for the
          group members.

        * For any other resources, the granted permission is object access
          permission for these resources.

        The User Role must be user-defined.

        HMC/SE version requirements:

        * HMC version >= 2.13.0

        Authorization requirements:

        * Task permission to the "Manage User Roles" task.

        Parameters:

          permitted_object (:class:`~zhmcclient.BaseResource` or :term:`string`):
            Permitted object(s), either as a Python resource object (e.g.
            :class:`~zhmcclient.Partition`), or as a resource class string (e.g.
            'partition').

            Must not be `None`.

          include_members (bool): Controls whether for Group resources, the
            operation applies additionally to its group member resources.

            If `None`, this parameter will be ignored. It should only be
            specified when the permitted object is a Group resource.

          view_only (bool): Controls whether for Task resources, the operation
            aplies to the view-only version of the task (if `True`), or to
            the full version of the task (if `False`). Only certain tasks
            support a view-only version.

            If `None`, this parameter will be ignored. It should only be
            specified when the permitted object is a Task resource.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """  # noqa: E501
        # pylint: enable=line-too-long
        if isinstance(permitted_object, BaseResource):
            perm_obj = permitted_object.uri
            perm_type = 'object'
        elif isinstance(permitted_object, str):
            perm_obj = permitted_object
            perm_type = 'object-class'
        else:
            raise TypeError(
                "permitted_object must be a string or BaseResource, but is: "
                f"{type(permitted_object)}")
        body = {
            'permitted-object': perm_obj,
            'permitted-object-type': perm_type,
        }
        if include_members is not None:
            body['include-members'] = include_members
        if view_only is not None:
            body['view-only-mode'] = view_only
        self.manager.session.post(
            self.uri + '/operations/remove-permission', resource=self,
            body=body)



# Copyright 2024 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _os_console module.
"""

from unittest import mock
import pytest
import websocket

from zhmcclient import OSConsole, Partition, OSConsoleConnectedError, \
    OSConsoleNotConnectedError

# pylint: disable=unused-import,line-too-long
from tests.common.http_mocked_fixtures import http_mocked_session  # noqa: F401
from tests.common.http_mocked_fixtures import http_mocked_cpc_dpm  # noqa: F401
from tests.common.http_mocked_fixtures import http_mocked_partition  # noqa: F401,E501
# pylint: enable=unused-import,line-too-long


def test_osc_initial_attrs(http_mocked_partition):  # noqa: F811
    # pylint: disable=redefined-outer-name, unused-argument
    """Test initial attributes of OSConsole."""

    ws_timeout = 42

    # Execute the code to be tested
    osc = OSConsole(partition=http_mocked_partition, ws_timeout=ws_timeout)

    assert osc.partition is http_mocked_partition
    assert osc.ws_timeout == ws_timeout

    assert osc.ws_uri is None
    assert osc.ws_full_uri is None
    assert osc.ws is None
    assert osc.previous_text is None

    assert osc.is_connected() is False


@mock.patch.object(Partition, 'create_os_websocket')
@mock.patch('websocket.WebSocket')
def test_osc_connect(
        websocket_mock, create_os_websocket_mock,
        http_mocked_partition):  # noqa: F811
    # pylint: disable=redefined-outer-name, unused-argument
    """All tests for OSConsole.connect/disconnect/is_connected()."""

    create_os_websocket_mock.return_value = '/api/websocket/fake-ws1'

    osc = OSConsole(partition=http_mocked_partition)

    # Test 1: Connect if not connected -> success
    osc.connect()
    assert osc.is_connected() is True

    # Test 2: Connect if connected -> fails
    with pytest.raises(OSConsoleConnectedError):
        osc.connect()
    assert osc.is_connected() is True

    # Test 3: Disconnect if connected -> success
    osc.disconnect()
    assert osc.is_connected() is False

    # Test 4: Disconnect if not connected -> fails
    with pytest.raises(OSConsoleNotConnectedError):
        osc.disconnect()
    assert osc.is_connected() is False


LPAR_OSC_RECV_ALL_TESTCASES = [
    # Testcases for test_osc_recv_all().
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - connected (bool): Whether initially to connect.
    # - recv_chunks (list): Data chunks to be received.
    # - exp_data (str): Expected received data.
    # - exp_exc_type (class): Expected exception type, or None for success.

    (
        "Not connected",
        False,
        ['line1\r\r\n'],
        None,
        OSConsoleNotConnectedError,
    ),
    (
        "No data",
        True,
        [],
        '',
        None,
    ),
    (
        "One chunk of data without NL",
        True,
        ['line1'],
        'line1',
        None,
    ),
    (
        "One chunk of data with NL",
        True,
        ['line1\n'],
        'line1\n',
        None,
    ),
    (
        "Two chunks of data with NL",
        True,
        ['line1\n', 'line2\n'],
        'line1\nline2\n',
        None,
    ),
    (
        "One chunk of data with one CR before NL",
        True,
        ['line1\r\n'],
        'line1\n',
        None,
    ),
    (
        "One chunk of data with two CRs before NL",
        True,
        ['line1\r\r\n'],
        'line1\n',
        None,
    ),
]


@pytest.mark.parametrize(
    "desc, connected, recv_chunks, exp_data, exp_exc_type",
    LPAR_OSC_RECV_ALL_TESTCASES)
@mock.patch.object(Partition, 'create_os_websocket')
@mock.patch('websocket.WebSocket')
def test_osc_recv_all(
        websocket_mock, create_os_websocket_mock,
        http_mocked_partition,  # noqa: F811
        desc, connected, recv_chunks, exp_data, exp_exc_type):
    # pylint: disable=redefined-outer-name, unused-argument
    """All tests for OSConsole.recv_all()."""

    create_os_websocket_mock.return_value = '/api/websocket/fake-ws1'
    osc = OSConsole(partition=http_mocked_partition, ws_timeout=1)

    if connected:
        osc.connect()

    if exp_exc_type:
        with pytest.raises(exp_exc_type):

            # The code to be tested
            osc.recv_all()

    else:

        # Prepare the data to be received
        last_chunk = websocket.WebSocketTimeoutException()
        osc.ws.recv.side_effect = recv_chunks + [last_chunk]

        # The code to be tested
        data = osc.recv_all()

        assert data == exp_data

        # pylint: disable=no-member
        assert osc.ws.recv.call_count == len(recv_chunks) + 1

    if osc.is_connected():
        osc.disconnect()


LPAR_OSC_RECV_LINE_TESTCASES = [
    # Testcases for test_osc_recv_line().
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - connected (bool): Whether initially to connect.
    # - recv_chunks (list): Data chunks available for receiving.
    # - exp_call_count (str): Expected number of calls to ws.recv()
    # - exp_data (str): Expected received data.
    # - exp_exc_type (class): Expected exception type, or None for success.

    (
        "Not connected",
        False,
        ['line1\n'],
        1,
        None,
        OSConsoleNotConnectedError,
    ),
    (
        "No data available",
        True,
        [],
        1,
        '',
        None,
    ),
    (
        "One chunk of data without NL available",
        True,
        ['line1'],
        2,
        'line1',
        None,
    ),
    (
        "One chunk of data with NL available",
        True,
        ['line1\n'],
        1,
        'line1\n',
        None,
    ),
    (
        "Two chunks of data with NL available",
        True,
        ['line1\n', 'line2\n'],
        1,
        'line1\n',
        None,
    ),
    (
        "One chunk of data with one CR before NL available",
        True,
        ['line1\r\n'],
        1,
        'line1\n',
        None,
    ),
    (
        "One chunk of data with two CRs before NL available",
        True,
        ['line1\r\r\n'],
        1,
        'line1\n',
        None,
    ),
    (
        "One chunk of data with two CRs before NL available",
        True,
        ['line1\r\r\n'],
        1,
        'line1\n',
        None,
    ),
    (
        "One chunk of data without NL and one chunk with NL available",
        True,
        ['line1', '\n'],
        2,
        'line1\n',
        None,
    ),
]


@pytest.mark.parametrize(
    "desc, connected, recv_chunks, exp_call_count, exp_data, exp_exc_type",
    LPAR_OSC_RECV_LINE_TESTCASES)
@mock.patch.object(Partition, 'create_os_websocket')
@mock.patch('websocket.WebSocket')
def test_osc_recv_line(
        websocket_mock, create_os_websocket_mock,
        http_mocked_partition,  # noqa: F811
        desc, connected, recv_chunks, exp_call_count, exp_data, exp_exc_type):
    # pylint: disable=redefined-outer-name, unused-argument
    """All tests for OSConsole.recv_line()."""

    create_os_websocket_mock.return_value = '/api/websocket/fake-ws1'
    osc = OSConsole(partition=http_mocked_partition, ws_timeout=1)

    if connected:
        osc.connect()

    if exp_exc_type:
        with pytest.raises(exp_exc_type):

            # The code to be tested
            osc.recv_line()

    else:

        # Prepare the data to be received
        last_chunk = websocket.WebSocketTimeoutException()
        osc.ws.recv.side_effect = recv_chunks + [last_chunk]

        # The code to be tested
        data = osc.recv_line()

        assert data == exp_data

        # pylint: disable=no-member
        assert osc.ws.recv.call_count == exp_call_count

    if osc.is_connected():
        osc.disconnect()


LPAR_OSC_SEND_TESTCASES = [
    # Testcases for test_osc_send().
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - connected (bool): Whether initially to connect.
    # - send_data (list): Data to be sent, or None.
    # - exp_exc_type (class): Expected exception type, or None for success.

    (
        "Not connected",
        False,
        None,
        OSConsoleNotConnectedError,
    ),
    (
        "No data",
        True,
        '',
        None,
    ),
    (
        "Data without NL",
        True,
        'line1',
        None,
    ),
    (
        "Data with one trailing NL",
        True,
        'line1\n',
        None,
    ),
    (
        "Data with two NLs",
        True,
        'line1\nline2\n',
        None,
    ),
]


@pytest.mark.parametrize(
    "desc, connected, send_data, exp_exc_type",
    LPAR_OSC_SEND_TESTCASES)
@mock.patch.object(Partition, 'create_os_websocket')
@mock.patch('websocket.WebSocket')
def test_osc_send(
        websocket_mock, create_os_websocket_mock,
        http_mocked_partition,  # noqa: F811
        desc, connected, send_data, exp_exc_type):
    # pylint: disable=redefined-outer-name, unused-argument
    """All tests for OSConsole.send()."""

    create_os_websocket_mock.return_value = '/api/websocket/fake-ws1'
    osc = OSConsole(partition=http_mocked_partition, ws_timeout=1)

    if connected:
        osc.connect()

    if exp_exc_type:
        with pytest.raises(exp_exc_type):

            # The code to be tested
            osc.send(send_data)

    else:

        # The code to be tested
        osc.send(send_data)

        # pylint: disable=no-member
        osc.ws.send_text.assert_called_once_with(send_data)

    if osc.is_connected():
        osc.disconnect()


# TODO: Tests for OSConsole.login_linux()


LPAR_OSC_EXEC_CMD_TESTCASES = [
    # Testcases for test_osc_exec_cmd().
    # The list items are tuples with the following items:
    # - desc (string): description of the testcase.
    # - connected (bool): Whether initially to connect.
    # - command (str): Command to be executed, or None.
    # - output (str): Command output to be received, or None.
    # - exp_exc_type (class): Expected exception type, or None for success.

    (
        "Not connected",
        False,
        'noout',
        None,
        OSConsoleNotConnectedError,
    ),
    (
        "Command with no output",
        True,
        'noout',
        '',
        None,
    ),
    (
        "Command with one line of output",
        True,
        'oneline',
        'line1\n',
        None,
    ),
    (
        "Command with two lines of output",
        True,
        'oneline',
        'line1\nline2\n',
        None,
    ),
]


@pytest.mark.parametrize(
    "desc, connected, command, output, exp_exc_type",
    LPAR_OSC_EXEC_CMD_TESTCASES)
@mock.patch.object(Partition, 'create_os_websocket')
@mock.patch('websocket.WebSocket')
def test_osc_exec_cmd(
        websocket_mock, create_os_websocket_mock,
        http_mocked_partition,  # noqa: F811
        desc, connected, command, output, exp_exc_type):
    # pylint: disable=redefined-outer-name, unused-argument
    """All tests for OSConsole.execute_command()."""

    create_os_websocket_mock.return_value = '/api/websocket/fake-ws1'
    osc = OSConsole(partition=http_mocked_partition, ws_timeout=1)

    if connected:
        osc.connect()

    if exp_exc_type:
        with pytest.raises(exp_exc_type):

            # The code to be tested
            osc.execute_command(command)

    else:

        # Prepare the command output
        last_chunk = websocket.WebSocketTimeoutException()
        osc.ws.recv.side_effect = [output, last_chunk]

        # The code to be tested
        act_output = osc.execute_command(command)

        assert act_output == output

        # pylint: disable=no-member
        osc.ws.send_text.assert_called_once_with(command + '\n')

    if osc.is_connected():
        osc.disconnect()



#!/usr/bin/env python
# Copyright 2016,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that lists CPCs with partial and full properties to show timing
difference.
"""

import sys
import requests.packages.urllib3
from datetime import datetime

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    for full_properties in (False, True):
        print(f"Listing CPCs with full_properties={full_properties} ...")
        start_dt = datetime.now()
        cpcs = client.cpcs.list(full_properties)
        end_dt = datetime.now()
        duration = end_dt - start_dt
        print(f"Duration: {duration}")
        for cpc in cpcs:
            print(f"Number of properties returned for CPC {cpc.name}: "
                  f"{len(cpc.properties)}")

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for the single console object.

These tests do not change the console object.
"""


from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import runtest_find_list, runtest_get_properties, \
    validate_list_features

urllib3.disable_warnings()

# Properties in minimalistic Console objects (e.g. find_by_name())
CONSOLE_MINIMAL_PROPS = ['object-uri']

# Properties in Console objects returned by list() without full props
CONSOLE_LIST_PROPS = ['object-uri']

# Properties whose values can change between retrievals of Console objects
CONSOLE_VOLATILE_PROPS = []


def test_console_find_list(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    client = zhmcclient.Client(hmc_session)

    # Pick the single console
    console = client.consoles.console

    runtest_find_list(
        hmc_session, client.consoles, console.name, 'name',
        'object-uri', CONSOLE_VOLATILE_PROPS, CONSOLE_MINIMAL_PROPS,
        CONSOLE_LIST_PROPS)


def test_console_property(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    client = zhmcclient.Client(hmc_session)

    # Select a property that is not returned by list()
    non_list_prop = 'description'

    runtest_get_properties(client.consoles, non_list_prop)


def test_console_list_features(hmc_session):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Tests Console.list_api_features() with and without passing a parameter.
    """
    client = zhmcclient.Client(hmc_session)
    console = client.consoles.console

    validate_list_features(client.query_api_version(),
                           console.list_api_features(),
                           console.list_api_features('cpc.*'))



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for storage groups (on CPCs in DPM mode).

These tests do not change any existing storage groups, but create, modify and
delete test storage groups.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    skipif_no_storage_mgmt_feature, runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic StorageGroup objects (e.g. find_by_name())
STOGRP_MINIMAL_PROPS = ['object-uri', 'name']

# Properties in StorageGroup objects returned by list() without full props
STOGRP_LIST_PROPS = ['object-uri', 'cpc-uri', 'name', 'fulfillment-state',
                     'type']

# Properties whose values can change between retrievals of StorageGroup objs
STOGRP_VOLATILE_PROPS = []


def test_stogrp_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        console = cpc.manager.client.consoles.console
        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the storage groups to test with
        stogrp_list = cpc.list_associated_storage_groups()
        if not stogrp_list:
            skip_warn(f"No storage groups associated to CPC {cpc.name} "
                      f"managed by HMC {hd.host}")
        stogrp_list = pick_test_resources(stogrp_list)

        for stogrp in stogrp_list:
            print(f"Testing on CPC {cpc.name} with storage group "
                  f"{stogrp.name!r}")
            runtest_find_list(
                session, console.storage_groups, stogrp.name, 'name',
                'object-uri', STOGRP_VOLATILE_PROPS, STOGRP_MINIMAL_PROPS,
                STOGRP_LIST_PROPS)


def test_stogrp_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the storage groups to test with
        stogrp_list = cpc.list_associated_storage_groups()
        if not stogrp_list:
            skip_warn(f"No storage groups associated to CPC {cpc.name} "
                      f"managed by HMC {hd.host}")
        stogrp_list = pick_test_resources(stogrp_list)

        for stogrp in stogrp_list:
            print(f"Testing on CPC {cpc.name} with storage group "
                  f"{stogrp.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(stogrp.manager, non_list_prop)


def test_stogrp_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a storage group.
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_no_storage_mgmt_feature(cpc)

        print(f"Testing on CPC {cpc.name}")

        console = cpc.manager.client.consoles.console
        stogrp_name = TEST_PREFIX + ' test_stogrp_crud stogrp1'
        stogrp_name_new = stogrp_name + ' new'

        # Ensure clean starting point
        try:
            stogrp = console.storage_groups.find(name=stogrp_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test storage group from previous run: "
                f"{stogrp_name!r} on CPC {cpc.name}", UserWarning)
            stogrp.delete()
        try:
            stogrp = console.storage_groups.find(name=stogrp_name_new)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                "Deleting test storage group from previous run: "
                f"{stogrp_name_new!r} on CPC {cpc.name}", UserWarning)
            stogrp.delete()

        # Test creating the storage group

        stogrp_input_props = {
            'cpc-uri': cpc.uri,
            'name': stogrp_name,
            'description': 'Test storage group for zhmcclient end2end tests',
            'type': 'fcp',
        }
        stogrp_auto_props = {
            'shared': True,
            'fulfillment-state': 'pending',
        }

        # The code to be tested
        stogrp = console.storage_groups.create(stogrp_input_props)

        for pn, exp_value in stogrp_input_props.items():
            assert stogrp.properties[pn] == exp_value, (
                f"Unexpected value for property {pn!r} of storage group:\n"
                f"{stogrp.properties!r}")
        stogrp.pull_full_properties()
        for pn, exp_value in stogrp_input_props.items():
            assert stogrp.properties[pn] == exp_value, (
                f"Unexpected value for property {pn!r} of storage group:\n"
                f"{stogrp.properties!r}")
        for pn, exp_value in stogrp_auto_props.items():
            assert pn in stogrp.properties, (
                f"Automatically returned property {pn!r} is not in "
                f"created storage group:\n{stogrp!r}")
            assert stogrp.properties[pn] == exp_value, (
                f"Unexpected value for property {pn!r} of storage group:\n"
                f"{stogrp.properties!r}")

        # Test updating a property of the storage group

        new_desc = "Updated storage group description."

        # The code to be tested
        stogrp.update_properties(dict(description=new_desc))

        assert stogrp.properties['description'] == new_desc
        stogrp.pull_full_properties()
        assert stogrp.properties['description'] == new_desc

        # Test renaming the storage group

        # The code to be tested
        stogrp.update_properties(dict(name=stogrp_name_new))

        assert stogrp.properties['name'] == stogrp_name_new
        stogrp.pull_full_properties()
        assert stogrp.properties['name'] == stogrp_name_new
        with pytest.raises(zhmcclient.NotFound):
            console.storage_groups.find(name=stogrp_name)

        # Test deleting the storage group

        # The code to be tested
        stogrp.delete()

        with pytest.raises(zhmcclient.NotFound):
            console.storage_groups.find(name=stogrp_name_new)
